#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import math

class Region(object):
    """
    Representation of a "region" in a simulation.

    Attributes:
        b: the beginning of the region
        e: the end of the region
        w: the "weight" assigned to the region
        l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation.

    This class is extended by:
        * :class:`fwdpy11.Sregion`
    """
    def __init__(self, beg, end, weight,coupled=True,label = 0):
        """
        Constructor

        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode:: 

            #A simple case
            import fwdpy11
            r = fwdpy11.Region(0,1,1)
            #A more "biological" case:
            #  The region covers positions 1 through 1,000,
            #  and the per-base pair "weight" is 1e-5:
            r = fwdpy11.Region(1,1000,1e-5,True)
        """
        if math.isinf(beg):
            raise ValueError("fwdpy11.Region: beg not finite")
        if math.isinf(end):
            raise ValueError("fwdpy11.Region: end not finite")
        if math.isinf(weight):
            raise ValueError("fwdpy11.Region: weight not finite")
        if math.isnan(beg):
            raise ValueError("fwdpy11.Region: beg not a number")
        if math.isnan(end):
            raise ValueError("fwdpy11.Region: end not a number")
        if math.isnan(weight):
            raise ValueError("fwdpy11.Region: weight not a number")
        if weight < 0.0:
            raise ValueError("fwdpy11.Region: weight < 0.0")
        self.b=float(beg)
        self.e=float(end)
        self.w=float(weight)
        self.c=coupled
        self.l=label
        if self.c is True:
            self.w = (self.e-self.b)*self.w
    def __str__(self):
        bstr="{:.9f}".format(self.b)
        estr="{:.9f}".format(self.e)
        wstr="{:.9f}".format(self.w)
        return "beg = " +bstr+ ", end = " + estr+ ", weight = "+wstr+", label = "+format(self.l)
        
class Sregion(Region):
    """
    Representation of a "region" in a simulation with a dominance term.

    This class is the base class for a general set of objects representing distributions of fitness effects.

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * h: the dominance term
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation.

    .. note:: This class cannot be used directly to parameterize a simulation.  Rather, you must used
       a derived type that specifies a distribution of fitness effects.  These types include:
       :class:`fwdpy11.fwdpy11.ConstantS`,
       :class:`fwdpy11.fwdpy11.UniformS`,
       :class:`fwdpy11.fwdpy11.ExpS`,
       :class:`fwdpy11.fwdpy11.GammaS`, and 
       :class:`fwdpy11.fwdpy11.GaussianS`
    """
    def __init__(self, beg, end, weight, h=1.0,coupled=True,label = 0):
        """
        Constructor

        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode:: 

            #A simple case
            import fwdpy11
            #Examples for models where the 3 genotype fitnesses are
            #1, 1+sh, and 1+2s, respectively
            recessive = fwdpy11.Sregion(0,1,1,0)
            additive = fwdpy11.Sregion(0,1,1,1.0)
            dominant = fwdpy11.Sregion(0,1,1,2.0)
        """
        if math.isinf(h):
            raise ValueError("fwdpy11.Sregion: h not finite")
        if math.isnan(h):
            raise ValueError("fwdpy11.Segion: h not a number")
        self.h=float(h)
        super(Sregion,self).__init__(beg,end,weight,coupled,label)
    def __str__(self):
        return "h = "+"{:.9f}".format(self.h)+", "+super(Sregion,self).__str__()
        
class GammaS(Sregion):
    """
    Gamma distribution of fitness effects

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * mean: mean of the Gamma
        * shape: shape of the Gamma
        * h: the dominance term
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation
    """
    def __init__(self, beg, end, weight, mean, shape, h=1.0,coupled=True,label = 0):
        """
        Constructor

        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param mean: the mean selection coefficient
        :param shape: the shape parameter of the distribution
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode:: 

            #A simple case
            import fwdpy11
            gdist = fwdpy11.GammaS(0,1,1,-0.1,0.35)
        """
        if math.isinf(mean):
            raise ValueError("fwdpy11.GammaS: mean not finite")
        if math.isinf(shape):
            raise ValueError("fwdpy11.GammaS: shape not finite")
        if math.isnan(mean):
            raise ValueError("fwdpy11.GammaS: mean not a number")
        if math.isnan(shape):
            raise ValueError("fwdpy11.GammaS: shape not a number")
        self.mean=float(mean)
        self.shape=float(shape)
        super(GammaS,self).__init__(beg,end,weight,h,coupled,label)
    def callback(self):
        from .fwdpp_extensions import makeGammaSH
        return makeGammaSH(self.mean,self.shape,self.h)
    def __str__(self):
        return "Gamma DFE, mean = "+"{:.9f}".format(self.mean)+", shape = "+"{:.9f}".format(self.shape)+", "+super(GammaS,self).__str__()
        
class ConstantS(Sregion):
    """
    Constant/fixed selection coefficient

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * s: the selection coefficient
        * h: the dominance term
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation
    """
    def __init__(self, beg, end, weight, s, h=1.0,coupled=True,label = 0):
        """
        Constructor

        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param s: the selection coefficient
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode::

            #A simple case
            import fwdpy11
            #s = -0.1 and h = 0
            constantS = fwdpy11.ConstantS(0,1,1,-0.1,0)
        """
        if math.isinf(s):
            raise ValueError("fwdpy11.ConstantS: s not finite")
        if math.isnan(s):
            raise ValueError("fwdpy11.ConstantS: s not a number")
        self.s=float(s)
        super(ConstantS,self).__init__(beg,end,weight,h,coupled,label)
    def callback(self):
        from .fwdpp_extensions import makeConstantSH 
        return makeConstantSH(self.s,self.h)
    def __str__(self):
        return "Constant s DFE, s = "+"{:.9f}".format(self.s)+", "+super(ConstantS,self).__str__()

class UniformS(Sregion):
    """
    Uniform distribution on selection coefficients

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * lo: the lower bound on s
        * hi: the upper bound on s
        * h: the dominance term
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation
    """
    def __init__(self, beg, end, weight, lo, hi, h=1.0,coupled=True):
        """
        Constructor
    
        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param lo: lower bound on s
        :param hi: upper bound on s
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode:: 

            #A simple case
            import fwdpy11
            #s is uniform on [0,-1]
            constantS = fwdpy11.UniformS(0,1,1,0,-1,0)
        """
        if math.isinf(lo):
            raise ValueError("fwdpy11.UniformS: lo not finite")
        if math.isinf(hi):
            raise ValueError("fwdpy11.UniformS: hi not finite")
        if math.isnan(lo):
            raise ValueError("fwdpy11.UniformS: lo not a number")
        if math.isnan(hi):
            raise ValueError("fwdpy11.UniformS: hi not a number")
        self.lo=float(lo)
        self.hi=float(hi)
        super(UniformS,self).__init__(beg,end,weight,h,coupled)
    def callback(self):
        from .fwdpp_extensions import makeUniformSH
        return makeUniformSH(self.lo,self.hi,self.h)
    def __str__(self):
        return "Uniform s DFE, lo = "+"{:.9f}".format(self.lo)+", hi = "+"{:.9f}".format(self.hi)+", "+super(UniformS,self).__str__()

class ExpS(Sregion):
    """
    Exponential distribution on selection coefficients

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * mean: the mean selection coefficient
        * h: the dominance term
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation
    """
    def __init__(self, beg, end, weight, mean, h=1.0,coupled=True,label = 0):
        """
        Constructor
    
        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param mean: the mean selection coefficient
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode::

            #A simple case
            import fwdpy11
            #s is exp(-0.1) and recessive
            constantS = fwdpy11.ExpS(0,1,1,0,-0.1,0)
        """
        if math.isinf(mean):
            raise ValueError("fwdpy11.ExpS: mean not finite")
        if math.isnan(mean):
            raise ValueError("fwdpy11.ExpS: mean not a number")
        self.mean=float(mean)
        super(ExpS,self).__init__(beg,end,weight,h,coupled,label)
    def callback(self):
        from .fwdpp_extensions import makeExpSH
        return makeExpSH(self.mean,self.h)
    def __str__(self):
        return "Exponential DFE, mean = "+"{:.9f}".format(self.mean)+", "+super(ExpS,self).__str__()

class GaussianS(Sregion):
    """
    Gaussian distribution on selection coefficients (effect sizes for sims of quantitative traits)

    Attributes:
        * b: the beginning of the region
        * e: the end of the region
        * w: the "weight" assigned to the region
        * sd: the standard deviation
        * h: the dominance ter
        * l: A label assigned to the region.  Labels must be integers, and can be used to 'tag' mutations arising in different regions.

    The mean is zero.

    See :func:`evolve_regions` for how this class may be used to parameterize a simulation
    """
    def __init__(self, beg, end, weight, sd, h=1.0,coupled=True,label = 0):
        """
        Constructor
    
        :param beg: the beginning of the region
        :param end: the end of the region
        :param weight: the weight to assign
        :param mean: the mean selection coefficient
        :param h: the dominance
        :param coupled: if True, the weight is converted to (end-beg)*weight
        :param label: Not relevant to recombining regions.  Otherwise, this value will be used to take mutations from this region.

        When coupled is True, the "weight" may be interpreted as a "per base pair"
        (or per unit, generally speaking) term.

        Example:

        .. testcode:: 

            #A simple case
            import fwdpy11
            #s N(0,0.1) and co-dominant
            constantS = fwdpy11.GaussianS(0,1,1,0,0.1,1)
        """
        if math.isinf(sd):
            raise ValueError("fwdpy11.GaussianS: sd not finite")
        if math.isnan(sd):
            raise ValueError("fwdpy11.GaussianS: sd not a number")
        self.sd=float(sd)
        super(GaussianS,self).__init__(beg,end,weight,h,coupled,label)
    def callback(self):
        from .fwdpp_extensions import makeGaussianSH
        return makeGaussianSH(self.sd,self.h)
    def __str__(self):
        return "Gaussian DFE, s.d. = "+"{:.9f}".format(self.sd)+", "+super(GaussianS,self).__str__()




from .wfevolve_qtrait import evolve_singlepop_regions_qtrait_cpp
def evolve_regions_sampler_fitness(rng,pop,popsizes,mu_neutral,
        mu_selected,recrate,nregions,sregions,recregions,trait_model,
        recorder,environments,selfing_rate = 0.):
    """
    Evolve a single deme according to a Wright-Fisher life cycle 
    with arbitrary changes in population size, a specified fitness model,
    and a temporal sampler.

    The fitness model is Gaussian stabilizing selection with respect to an
    optimum:

    .. math::

        w = e^{-\\frac{(P-O)^2}{2VS}}

    The trait value is :math:`P=G+E`, where :math:`G` is the genetic component, calculated
    using an object of type `fwdpy11.fitness.SpopFitness`.  The "environmental" term, 
    :math:`E` is a Gassian deviate with mean 0 and standard deviation 
    :math:`\sigma_e`.

    :param rng: A :class:`fwdpy11.fwdpy11_types.GSLrng`
    :param pop: A :class:`fwdpy11.fwdpy11_types.Spop`
    :param popsizes: A 1d NumPy array representing population sizes over time.
    :param mu_neutral: The neutral mutation rate (per gamete, per generation)
    :param mu_selected: The selected mutation rate (per gamete, per generation)
    :param recrate: The recombination reate (per diploid, per generation)
    :param nregions: A list of :class:`fwdpy11.regions.Region`.
    :param sregions: A list of :class:`fwdpy11.regions.Sregion`.
    :param recregions: A list of :class:`fwdpy11.regions.Region`.
    :param trait_model: A :class:`fwdpy11.fitness.SpopFitness`.
    :param recorder: A callable to record data from the population.
    :param environments: A list of environmental conditions (see below).
    :param selfing_rate: (default 0.0) The probability than an individual selfs.

    The environmental conditions are specified by a list of tuples containing
    four elements each.  These elements are:

    * Generation.  
    * optimum, the optimum trait value.
    * VS, the intensity of selection against deviations from the optimum trait value.
    * :math:`\sigma_e`, the standard deviation of random noise added to trait value.

    When the population's generation first becomes :math:`\geq` `Generation`, then 
    the values for `optimum`, `VS` and :math:`\sigma_e` are applied.
    """
    from .internal import makeMutationRegions,makeRecombinationRegions
    mm=makeMutationRegions(nregions,sregions)
    rm=makeRecombinationRegions(recregions)
    evolve_singlepop_regions_qtrait_cpp(rng,pop,popsizes,mu_neutral,
            mu_selected,recrate,mm,rm,trait_model,recorder,selfing_rate,environments)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from .wfevolve import evolve_singlepop_regions_cpp

def quick_sim(ngens = None):
    """
    A convenience function for rapidly creating a
    :class:`fwdpy11.fwdpy11_types.Spop`

    .. testcode:: 

        import fwdpy11.wright_fisher
        #This will simulate N=1e3 for 10N generations
        pop = fwdpy11.wright_fisher.quick_sim()
        print(pop.N)
        print(pop.generation)

    The output is:

    .. testoutput::

      1000
      10000

    .. note::
        Implemented via a call to :func:`fwdpy11.wright_fisher.evolve`
    """
    from .fwdpy11_types import GSLrng,Spop
    rng = GSLrng(42)
    pop=Spop(1000)
    if ngens is None:
        evolve(rng,pop)
    else:
        import numpy as np
        nlist = np.array([pop.N]*ngens,dtype=np.uint32)
        evolve(rng,pop,nlist)
    return pop

def evolve(rng,pop,popsizes = None,mu_neutral=None,
        mu_selected = None,recrate=None,sregions=None):
    """
    
    .. testcode::

        import fwdpy11 as fp11
        import fwdpy11.wright_fisher as wf
        import numpy as np
        rng = fp11.GSLrng(42)
        p = fp11.Spop(1000)
        wf.evolve(rng,p)
        print(p.generation)
        nlist=np.array([5000]*323,dtype=np.uint32)
        wf.evolve(rng,p,nlist)
        print(p.N)
        print(p.generation)
            
    .. testoutput::

      10000
      5000
      10323

    """
    if popsizes is None:
        import numpy as np
        popsizes = np.array([pop.N]*10*pop.N,dtype=np.uint32)
    if mu_neutral is None:
        mu_neutral = 100./float(4*pop.N)
    if mu_selected is None:
        mu_selected = 10./float(4*pop.N)
    if recrate is None:
        recrate = 100./float(4*pop.N)
    if sregions is None:
        from .regions import ExpS
        sregions = [ExpS(0,1,1,-0.1,1.0)]
    from .regions import Region
    nr=[Region(0,1,1)]
    return evolve_regions(rng,pop,popsizes,mu_neutral,
            mu_selected,recrate,nr,sregions,
            nr)

def evolve_regions(rng,pop,popsizes,mu_neutral,
        mu_selected,recrate,nregions,sregions,recregions,
        selfing_rate = 0.):
    """
    Evolve a single deme according to a Wright-Fisher life cycle 
    with arbitrary changes in population size and a temporal sampler.

    :param rng: A :class:`fwdpy11.fwdpy11_types.GSLrng`
    :param pop: A :class:`fwdpy11.fwdpy11_types.Spop`
    :param popsizes: A 1d NumPy array representing population sizes over time.
    :param mu_neutral: The neutral mutation rate (per gamete, per generation)
    :param mu_selected: The selected mutation rate (per gamete, per generation)
    :param recrate: The recombination reate (per diploid, per generation)
    :param nregions: A list of :class:`fwdpy11.regions.Region`.
    :param sregions: A list of :class:`fwdpy11.regions.Sregion`.
    :param recregions: A list of :class:`fwdpy11.regions.Region`.
    :param recorder: A callable to record data from the population.
    :param selfing_rate: (default 0.0) The probability than an individual selfs.

    .. note:: 
        The fitness model will be :class:`fwdpy11.fitness.SpopAdditive` constructed
        with a scaling of 2.0. This function calls 
        :func:`fwdpy11.wright_fisher.evolve_regions_sampler`, passing in a
        :class:`fwdpy11.temporal_samplers.RecordNothing` object.
    """
    from .temporal_samplers import RecordNothing
    recorder=RecordNothing()
    return evolve_regions_sampler(rng,pop,popsizes,mu_neutral,
            mu_selected,recrate,nregions,sregions,recregions,
            recorder,selfing_rate)

def evolve_regions_sampler(rng,pop,popsizes,mu_neutral,
        mu_selected,recrate,nregions,sregions,recregions,
        recorder,selfing_rate = 0.):
    """
    Evolve a single deme according to a Wright-Fisher life cycle 
    with arbitrary changes in population size and a temporal sampler.

    :param rng: A :class:`fwdpy11.fwdpy11_types.GSLrng`
    :param pop: A :class:`fwdpy11.fwdpy11_types.Spop`
    :param popsizes: A 1d NumPy array representing population sizes over time.
    :param mu_neutral: The neutral mutation rate (per gamete, per generation)
    :param mu_selected: The selected mutation rate (per gamete, per generation)
    :param recrate: The recombination reate (per diploid, per generation)
    :param nregions: A list of :class:`fwdpy11.regions.Region`.
    :param sregions: A list of :class:`fwdpy11.regions.Sregion`.
    :param recregions: A list of :class:`fwdpy11.regions.Region`.
    :param recorder: A callable to record data from the population.
    :param selfing_rate: (default 0.0) The probability than an individual selfs.

    .. note:: 
        The fitness model will be :class:`fwdpy11.fitness.SpopAdditive` constructed
        with a scaling of 2.0.
    """
    from .fitness import SpopAdditive
    fitness = SpopAdditive(2.0)
    return evolve_regions_sampler_fitness(rng,pop,popsizes,mu_neutral,
            mu_selected,recrate,nregions,sregions,recregions,fitness,
            recorder,selfing_rate)

def evolve_regions_sampler_fitness(rng,pop,popsizes,mu_neutral,
        mu_selected,recrate,nregions,sregions,recregions,fitness,
        recorder,selfing_rate = 0.):
    """
    Evolve a single deme according to a Wright-Fisher life cycle 
    with arbitrary changes in population size, a specified fitness model,
    and a temporal sampler.

    :param rng: A :class:`fwdpy11.fwdpy11_types.GSLrng`
    :param pop: A :class:`fwdpy11.fwdpy11_types.Spop`
    :param popsizes: A 1d NumPy array representing population sizes over time.
    :param mu_neutral: The neutral mutation rate (per gamete, per generation)
    :param mu_selected: The selected mutation rate (per gamete, per generation)
    :param recrate: The recombination reate (per diploid, per generation)
    :param nregions: A list of :class:`fwdpy11.regions.Region`.
    :param sregions: A list of :class:`fwdpy11.regions.Sregion`.
    :param recregions: A list of :class:`fwdpy11.regions.Region`.
    :param fitness: A :class:`fwdpy11.fitness.SpopFitness`.
    :param recorder: A callable to record data from the population.
    :param selfing_rate: (default 0.0) The probability than an individual selfs.
    """
    from .internal import makeMutationRegions,makeRecombinationRegions
    mm=makeMutationRegions(nregions,sregions)
    rm=makeRecombinationRegions(recregions)
    evolve_singlepop_regions_cpp(rng,pop,popsizes,mu_neutral,
            mu_selected,recrate,mm,rm,fitness,recorder,selfing_rate)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
def makeMutationRegions(neutral,selected):
    """
    Convert user input into :class:`~fwdpy11.fwdpp_extensions.MutationRegions`

    :param neutral: A list of :class:`fwdpy11.regions.Region` objects.
    :param selected: A list of :class:`fwdpy11.regions.Sregion` objects.

    :rtype: :class:`fwdpy11.fwdpp_extensions.MutationRegions`

    .. note:: Used by various "evolve" functions.  Users probably won't need to call this.
    
        >>> import fwdpy11 as fp11
        >>> nregions = [fp11.Region(0,0.5,1),fp11.Region(1,1.5,1)]
        >>> sregions = [fp11.ExpS(0,0.5,1,1),fp11.GaussianS(1,1.5,1,0.25)]
        >>> mr = fp11.makeMutationRegions(nregions,sregions)
        >>> type(mr)
        <class 'fwdpy11.fwdpp_extensions.MutationRegions'>

    One or both lists may be empty:

        >>> mr = fp11.makeMutationRegions([],[])

    Neither list may be None:

        >>> mr = fp11.makeMutationRegions([],None)
        Traceback (most recent call last):
            ...
        TypeError: 'NoneType' object is not iterable
    """
    nbeg = [i.b for i in neutral]
    nend = [i.e for i in neutral]
    nweights = [i.w for i in neutral]
    sbeg = [i.b for i in selected]
    send = [i.e for i in selected]
    sweights = [i.w for i in selected]
    sh = [i.callback() for i in selected]
    from .fwdpp_extensions import MutationRegions
    return MutationRegions(nbeg,nend,nweights,sbeg,send,sweights,sh)

def makeRecombinationRegions(regions):
    """
    Convert user input into :class:`~fwdpy11.fwdpp_extensions.RecombinationRegions`

    :param neutral: A list of :class:`fwdpy11.regions.Region` objects.
    :param selected: A list of :class:`fwdpy11.regions.Sregion` objects.

    :rtype: :class:`fwdpy11.fwdpp_extensions.MutationRegions`

    .. note:: Used by various "evolve" functions.  Users probably won't need to call this.
    
        >>> import fwdpy11 as fp11
        >>> rregions = [fp11.Region(0,0.5,1),fp11.Region(1,1.5,1)]
        >>> rr = fp11.makeRecombinationRegions(rregions)
        >>> type(rr)
        <class 'fwdpy11.fwdpp_extensions.RecombinationRegions'>

    """
    beg = [i.b for i in regions]
    end = [i.e for i in regions]
    weights = [i.w for i in regions]
    from .fwdpp_extensions import RecombinationRegions
    return RecombinationRegions(beg,end,weights)



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import setuptools
import os,glob

__version__ = '0.1-post3'


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)

PKGS=['fwdpy11']

INCLUDES=[
    'fwdpy11/headers',
    'fwdpy11/headers/fwdpp',
    # Path to pybind11 headers
    get_pybind_include(),
    get_pybind_include(user=True),
    os.path.join(sys.prefix, 'include')
]

LIBRARY_DIRS=[
    os.path.join(sys.prefix, 'lib')
    ]
ext_modules = [
    Extension(
        'fwdpy11.fwdpp_types',
        ['fwdpy11/src/fwdpp_types.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.fwdpp_extensions',
        ['fwdpy11/src/fwdpp_extensions.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.fwdpy11_types',
        ['fwdpy11/src/fwdpy11_types.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.sampling',
        ['fwdpy11/src/fwdpy11_sampling.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.fitness',
        ['fwdpy11/src/fwdpy11_fitness.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.trait_values',
        ['fwdpy11/src/fwdpy11_trait_values.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
        Extension(
        'fwdpy11.wfevolve',
        ['fwdpy11/src/wfevolve.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    Extension(
        'fwdpy11.wfevolve_qtrait',
        ['fwdpy11/src/wfevolve_qtrait.cc'],
        library_dirs=LIBRARY_DIRS,
        include_dirs=INCLUDES,
        libraries=['gsl','gslcblas'],
        language='c++'
    ),
    ]
    


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

#    if sys.platform == 'darwin':
#        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):# and sys.platform != 'darwin':
                opts.append('-fvisibility=hidden')
            if has_flag(self.compiler,'-g0'):
                opts.append('-g0')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

#Figure out the headers we need to install:
generated_package_data={}
for root, dirnames, filenames in os.walk('fwdpy11/headers'):
    if 'testsuite' not in root and 'examples' not in root and 'python_examples' not in root:
        g=glob.glob(root+'/*.hh')
        if len(g)>0:
            replace=root.replace('/','.')
            if replace not in PKGS:
                PKGS.append(replace) #If there's a header file, we add the directory as a package
            generated_package_data[replace]=['*.hh']
        g=glob.glob(root+'/*.hpp')
        if len(g)>0:
            replace=root.replace('/','.')
            if replace not in PKGS:
                PKGS.append(replace) #If there's a header file, we add the directory as a package
            try:
                if '*.hpp' not in generated_package_data[replace]:
                    generated_package_data[replace].append('*.hpp')
            except:
                generated_package_data[replace]=['*.hpp']
        g=glob.glob(root+'/*.tcc')
        if len(g)>0:
            replace=root.replace('/','.')
            if replace not in PKGS:
                PKGS.append(replace) #If there's a template implementation file, we add the directory as a package
            try:
                if '*.tcc' not in generated_package_data[replace]:
                    generated_package_data[replace].append('*.tcc')
            except:
                generated_package_data[replace]=['*.tcc']

long_desc = open("README.rst").read()

setup(
    name='fwdpy11',
    version=__version__,
    author='Kevin Thornton',
    author_email='krthornt@uci.edu',
    url='http://molpopgen.github.io/fwdpy11',
    classifiers=['Intended Audience :: Science/Research',
               'Topic :: Scientific/Engineering :: Bio-Informatics',
               'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'],
    description='Forward-time population genetic simulation in Python',
    license='GPL >= 3',
    requires=['pybind11','numpy'],
    provides=['fwdpy11'],
    obsoletes=['none'],
    data_files=[('fwdpy11',['COPYING', 'README.rst'])],
    long_description=long_desc,
    ext_modules=ext_modules,
    install_requires=['pybind11>=1.7'],
    cmdclass={'build_ext': BuildExt},
    packages=PKGS,
    package_data=generated_package_data,
    zip_safe=False,
)
